/*
 * Copyright © 2018 Apple Inc. and the ServiceTalk project authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.servicetalk.concurrent.api;

import io.servicetalk.concurrent.Cancellable;
import io.servicetalk.concurrent.SingleSource.Subscriber;

import java.util.function.Consumer;
import java.util.function.Supplier;
import javax.annotation.Nullable;

import static java.util.Objects.requireNonNull;

final class SingleDoOnUtils {
    private SingleDoOnUtils() {
        // No instances.
    }

    static <X> Supplier<Subscriber<? super X>> doOnSubscribeSupplier(Consumer<Cancellable> onSubscribe) {
        requireNonNull(onSubscribe);
        Subscriber<X> subscriber = new Subscriber<X>() {
            @Override
            public void onSubscribe(Cancellable cancellable) {
                onSubscribe.accept(cancellable);
            }

            @Override
            public void onSuccess(@Nullable X result) {
                // NOOP
            }

            @Override
            public void onError(Throwable t) {
                // NOOP
            }
        };
        return () -> subscriber;
    }

    static <X> Supplier<Subscriber<? super X>> doOnSuccessSupplier(Consumer<X> onSuccess) {
        requireNonNull(onSuccess);
        Subscriber<X> subscriber = new Subscriber<X>() {
            @Override
            public void onSubscribe(Cancellable cancellable) {
                // NOOP
            }

            @Override
            public void onSuccess(@Nullable X result) {
                onSuccess.accept(result);
            }

            @Override
            public void onError(Throwable t) {
                // NOOP
            }
        };
        return () -> subscriber;
    }

    static <X> Supplier<Subscriber<? super X>> doOnErrorSupplier(Consumer<Throwable> onError) {
        requireNonNull(onError);
        Subscriber<X> subscriber = new Subscriber<X>() {
            @Override
            public void onSubscribe(Cancellable cancellable) {
                // NOOP
            }

            @Override
            public void onSuccess(@Nullable X result) {
                // NOOP
            }

            @Override
            public void onError(Throwable t) {
                onError.accept(t);
            }
        };
        return () -> subscriber;
    }
}
